@@ -229,6 +229,8 @@ static void _print_scrub_ss(struct scrub_stats *ss)
 {
 	char t[4096];
 	struct tm tm;
+	time_t seconds;
+	unsigned hours;
 
 	if (!ss || !ss->t_start) {
 		printf("\tno stats available\n");
@@ -245,18 +247,21 @@ static void _print_scrub_ss(struct scrub_stats *ss)
 		t[sizeof(t) - 1] = '\0';
 		printf("\tscrub started at %s", t);
 	}
+
+	seconds = ss->duration;
+	hours = ss->duration / (60 * 60);
+	gmtime_r(&seconds, &tm);
+	strftime(t, sizeof(t), "%M:%S", &tm);
 	if (ss->finished && !ss->canceled) {
-		printf(" and finished after %llu seconds\n",
-		       ss->duration);
+		printf(" and finished after %02u:%s\n", hours, t);
 	} else if (ss->canceled) {
-		printf(" and was aborted after %llu seconds\n",
-		       ss->duration);
+		printf(" and was aborted after %02u:%s\n", hours, t);
 	} else {
 		if (ss->in_progress)
-			printf(", running for %llu seconds\n", ss->duration);
+			printf(", running for %02u:%s\n", hours, t);
 		else
-			printf(", interrupted after %llu seconds, not running\n",
-					ss->duration);
+			printf(", interrupted after %02u:%s, not running\n",
+					hours, t);
 	}
 }
 
@@ -385,7 +390,7 @@ static int scrub_open_file(const char *datafile, int m)
 static int scrub_open_file_r(const char *fn_base, const char *fn_local)
 {
 	int ret;
-	char datafile[BTRFS_PATH_NAME_MAX + 1];
+	char datafile[PATH_MAX];
 	ret = scrub_datafile(fn_base, fn_local, NULL,
 				datafile, sizeof(datafile));
 	if (ret < 0)
@@ -397,7 +402,7 @@ static int scrub_open_file_w(const char *fn_base, const char *fn_local,
 				const char *tmp)
 {
 	int ret;
-	char datafile[BTRFS_PATH_NAME_MAX + 1];
+	char datafile[PATH_MAX];
 	ret = scrub_datafile(fn_base, fn_local, tmp,
 				datafile, sizeof(datafile));
 	if (ret < 0)
@@ -409,8 +414,8 @@ static int scrub_rename_file(const char *fn_base, const char *fn_local,
 				const char *tmp)
 {
 	int ret;
-	char datafile_old[BTRFS_PATH_NAME_MAX + 1];
-	char datafile_new[BTRFS_PATH_NAME_MAX + 1];
+	char datafile_old[PATH_MAX];
+	char datafile_new[PATH_MAX];
 	ret = scrub_datafile(fn_base, fn_local, tmp,
 				datafile_old, sizeof(datafile_old));
 	if (ret < 0)
@@ -1073,15 +1078,11 @@ static int is_scrub_running_in_kernel(int fd,
 		memset(&sp, 0, sizeof(sp));
 		sp.scrub_args.devid = di_args[i].devid;
 		ret = ioctl(fd, BTRFS_IOC_SCRUB_PROGRESS, &sp.scrub_args);
-		if (ret < 0 && errno == ENODEV)
-			continue;
-		if (ret < 0 && errno == ENOTCONN)
-			return 0;
 		if (!ret)
 			return 1;
 	}
 
-	return 1;
+	return 0;
 }
 
 static const char * const cmd_scrub_start_usage[];
@@ -1127,7 +1128,7 @@ static int scrub_start(int argc, char **argv, int resume)
 	struct scrub_file_record *last_scrub = NULL;
 	char *datafile = strdup(SCRUB_DATA_FILE);
 	char fsid[BTRFS_UUID_UNPARSED_SIZE];
-	char sock_path[BTRFS_PATH_NAME_MAX + 1] = "";
+	char sock_path[PATH_MAX] = "";
 	struct scrub_progress_cycle spc;
 	pthread_mutex_t spc_write_mutex = PTHREAD_MUTEX_INITIALIZER;
 	void *terr;
@@ -1803,8 +1804,11 @@ out:
 	return !!err;
 }
 
+static const char scrub_cmd_group_info[] =
+"verify checksums of data and metadata";
+
 const struct cmd_group scrub_cmd_group = {
-	scrub_cmd_group_usage, NULL, {
+	scrub_cmd_group_usage, scrub_cmd_group_info, {
 		{ "start", cmd_scrub_start, cmd_scrub_start_usage, NULL, 0 },
 		{ "cancel", cmd_scrub_cancel, cmd_scrub_cancel_usage, NULL, 0 },
 		{ "resume", cmd_scrub_resume, cmd_scrub_resume_usage, NULL, 0 },
